from django.db import models
from django.apps import apps
from django.utils.html import format_html
from .utils import KPIService
from django_icon_picker.field import IconField
class KPI(models.Model):
    """
    Main KPI model with comparison conditions
    """

    name = models.CharField(max_length=100)
    description = models.TextField(blank=True, help_text="Optional description of the KPI")
    model_field = models.CharField(
        max_length=100,
        verbose_name="Model"
    )

    def __str__(self):
        return self.model_field
    
    
    def get_queryset(self, condition, target_field, target_value):
        """Get the queryset based on the condition and target"""
        app, model_name = self.model_field.split(".")
        queryset = apps.get_model(app, model_name).objects
        helper = KPIService()
        return helper.apply_condition(queryset, condition, target_field, target_value)
    

    class Meta:
        verbose_name = "KPI"
        verbose_name_plural = "KPIs"


class KpiCard(models.Model):
    """
    Card visualization for KPI with one-to-one relationship
    """
    TEXT_CONDITION_CHOICES = [
        ('EXACT', 'Exactly matches'),
        ('CONTAINS', 'Contains'),
        ('NOT_EXACT', 'Does not exactly match'),
        ('NOT_CONTAINS', 'Does not contain'),
    ]
    NUM_CONDITION_CHOICES = [
        ('GT', 'Greater Than'),
        ('LT', 'Less Than'),
        ('EQ', 'Equal'),
        ('GTE', 'Greater Than or Equal'),
        ('LTE', 'Less Than or Equal'),
        ('BETWEEN', 'Between'),
    ]
    CONDITION_CHOICES = [
        ('TEXT', TEXT_CONDITION_CHOICES),
        ('NUM', NUM_CONDITION_CHOICES),
        ('NONE', 'None'),
    ]
    OPERATION_CHOICES = [
        ('count', 'Count'),
        ('count_distinct', 'Count Distinct'),
        ('sum', 'Sum'),
        ('avg', 'Average'),
        ('min', 'Minimum'),
        ('max', 'Maximum'),
    ]
    kpi = models.ForeignKey(KPI, on_delete=models.CASCADE, related_name='card')
    name = models.CharField(max_length=100, help_text="Name of the card")
    description = models.TextField(blank=True, help_text="Optional description of the card")
    icon = IconField(max_length=50, help_text="Icon class or name")
    value_suffix = models.CharField(max_length=50, blank=True, help_text="Suffix to append to the value (e.g., %, $)")
    operation = models.CharField(max_length=16, choices=OPERATION_CHOICES, default='count')
    target_type = models.CharField(
        max_length=20,
        default='NUMBER',
        help_text="Type of the target value"
    )
    target_field = models.CharField(
        max_length=100,
        blank=True,
        help_text="Field to compare against target value"
    )
    condition = models.CharField(
        max_length=20, 
        choices=CONDITION_CHOICES,
        default='EXACT'
    )
    target_value = models.CharField(
        max_length=255,
        null=True,
        blank=True,
        help_text="Target value to achieve"
    )

    def svg_icon(self):
        return format_html(
            '<img src="{}" height="30" width="30"/>'.format(
                f"/{self.icon}"
                if self.icon.endswith(".svg")
                else f"https://api.iconify.design/{self.icon}.svg"
            )
        )

    @property
    def value(self):
        queryset = self.kpi.get_queryset(self.condition, self.target_field, self.target_value)

        if self.operation == 'count':
            return queryset.count()
        elif self.operation == 'count_distinct':
            return queryset.values(self.target_field).distinct().count()  
        elif self.operation == 'sum':
            return queryset.aggregate(models.Sum(self.target_field)).get(f'{self.target_field}__sum', 0) or 0
        elif self.operation == 'average':
            return queryset.aggregate(models.Avg(self.target_field)).get(f'{self.target_field}__avg', 0) or 0
        elif self.operation == 'min':
            return queryset.aggregate(models.Min(self.target_field)).get(f'{self.target_field}__min', 0) or 0
        elif self.operation == 'max':
            return queryset.aggregate(models.Max(self.target_field)).get(f'{self.target_field}__max', 0) or 0


    def __str__(self):
        return f"Card for {self.kpi.name}"


# class Table(models.Model):
#     """
#     Table visualization for KPI
#     """
#     kpi = models.OneToOneField(KPI, on_delete=models.CASCADE, related_name='table')
#     fields = models.JSONField(help_text="List of fields to display in table")
#     sort_by = models.CharField(max_length=100)
#     page_size = models.IntegerField(default=10)
#     is_paginated = models.BooleanField(default=True)

#     def __str__(self):
#         return f"Table for {self.kpi.name}"


# class Tile(models.Model):
#     """
#     Tile visualization for KPI
#     """
#     kpi = models.OneToOneField(KPI, on_delete=models.CASCADE, related_name='tile')
#     fields = models.JSONField(help_text="List of fields to display in tile")
#     layout = models.CharField(max_length=50, default='grid')
#     columns = models.IntegerField(default=3)

#     def __str__(self):
#         return f"Tile for {self.kpi.name}"


# class Chart(models.Model):
#     """
#     Chart visualization for KPI
#     """
#     CHART_TYPES = [
#         ('line', 'Line Chart'),
#         ('bar', 'Bar Chart'),
#         ('pie', 'Pie Chart'),
#         ('doughnut', 'Doughnut Chart'),
#         ('area', 'Area Chart'),
#     ]

#     kpi = models.OneToOneField(KPI, on_delete=models.CASCADE, related_name='chart')
#     chart_type = models.CharField(max_length=20, choices=CHART_TYPES)
#     field = models.CharField(max_length=100)
#     time_range = models.CharField(max_length=50, default='last_30_days')
#     show_legend = models.BooleanField(default=True)

#     def __str__(self):
#         return f"Chart for {self.kpi.name}"


# class Progress(models.Model):
#     """
#     Progress visualization for KPI
#     """
#     kpi = models.OneToOneField(KPI, on_delete=models.CASCADE, related_name='progress')
#     fields = models.JSONField(help_text="List of fields to track progress")
#     target_value = models.FloatField()
#     min_value = models.FloatField(default=0)
#     max_value = models.FloatField(default=100)
#     display_percentage = models.BooleanField(default=True)
#     color_scheme = models.CharField(max_length=50, default='default')

#     def __str__(self):
#         return f"Progress for {self.kpi.name}"


from django.urls import path
from . import views

urlpatterns = [
    path('get-model-fields/', views.get_model_fields, name='get_model_fields'),
    path('get-field-values/', views.get_field_values, name='get_field_values'),
]


from django.contrib import admin
from .models import KPI, KpiCard
from .forms import KPIAdminForm, CardAdminForm

@admin.register(KPI)
class KPIAdmin(admin.ModelAdmin):
    form = KPIAdminForm
    list_display = ('name', 'model_field')
    list_filter = ('model_field',)
    search_fields = ('name',)

@admin.register(KpiCard)
class CardAdmin(admin.ModelAdmin):
    form = CardAdminForm
    list_display = ['svg_icon', 'name', 'kpi', 'operation', 'target_field', 'condition', 'target_value', 'result']
    list_filter = ['kpi', 'operation', 'condition']
    search_fields = ['name', 'kpi__name', 'description']
    fieldsets = (
        (None, {'fields': ('kpi', 'name', 'description', 'icon')}),
        ('Value Settings', {'fields': ('value_suffix', 'operation')}),
        ('Target Settings', {'fields': ('target_type', 'target_field', 'condition', 'target_value')}),
    )

    def result(self, instance: KpiCard):
        return instance.value
    
    class Media:
        js = (
            'js/kpi_admin.js',
        )


# Suggested code may be subject to a license. Learn more: ~LicenseLog:4141344616.
from django import forms

from .utils import KPIService
from .models import KPI, KpiCard
# from core.widgets import IconPicker
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import gettext_lazy as _

class KPIAdminForm(forms.ModelForm):
    """
    Custom form for KPIAdmin
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Set choices dynamically in form initialization
        self.fields['model_field'] = forms.ChoiceField(
            choices=[('', '--- Select Model ---')] + KPIService.get_available_models(),
            required=True,
            widget=forms.Select
        )

    class Meta:
        model = KPI
        fields = '__all__'

class CardAdminForm(forms.ModelForm):
    """
    Custom form for CardAdmin
    """
    # icon = forms.CharField(widget=IconPicker())
        
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['target_type'].widget = forms.HiddenInput()
        self.fields['target_field'] = forms.CharField(
            required=True,
            widget=forms.Select(choices=[(self.instance.target_field, self.instance.target_field)],),
        )
        self.fields['target_value'] = forms.CharField(
            required=False,
            widget=forms.Select(choices=[(self.instance.target_value, self.instance.target_value)],),
        )
        self.fields["icon"].widget.attrs["model"] = self._meta.model.__name__
        if self.instance and self.instance.pk:
            object_id = self.instance.pk
            self.fields["icon"].widget.attrs["objectid"] = object_id
        else:
            last_item_id = (
                KpiCard.objects.last().id if KpiCard.objects.exists() else 1
            )
            next_id = last_item_id + 1
            self.fields["icon"].widget.attrs["objectid"] = next_id

    def clean(self):
        cleaned_data = super().clean()
        target_type = cleaned_data.get('target_type')
        target_value = cleaned_data.get('target_value')
        operation = cleaned_data.get('operation')
        condition = cleaned_data.get('condition')

        # Validate that target_value is present if condition is not NONE
        if condition != "NONE" and not target_value:
            raise ValidationError(_("Target value is required when a condition is selected."))

        # Validate operation against target_type
        if target_type and operation:
            is_numeric_field = issubclass(getattr(models, target_type), (models.IntegerField, models.FloatField, models.DecimalField))
            if not is_numeric_field and operation in ['sum', 'avg', 'max', 'min']:
                raise ValidationError(_(
                    f"Invalid operation '{operation}' for field type '{target_type}'. "
                    f"Choose 'count' or 'count_distinct' for non-numeric fields."
                ))
            
        # Validate condition against target_type 
        if target_type and condition.lower() in ['gt', 'gte', 'lt', 'lte', 'eq', 'between']:
            is_numeric_or_datetime_field = issubclass(
                getattr(models, target_type), 
                (models.IntegerField, models.FloatField, models.DecimalField, models.DateTimeField)
            )
            if not is_numeric_or_datetime_field:
                raise ValidationError(_(f"Comparison operations are not supported for '{target_type}' fields."))

        return cleaned_data
        

    class Meta:
        model = KpiCard
        fields = '__all__'


# utils.py
from django.apps import apps
from django.db.utils import OperationalError
from django.core.exceptions import AppRegistryNotReady
from datetime import datetime

class KPIService:
    # TODO: Handle excepts using field type
    """
    Helper class to handle KPI query conditions
    """
    
    def get_available_models():
        """
        Get all available models in the project excluding some built-in models
        Returns a list of tuples (model_name, verbose_name)
        """
        try:
            excluded_apps = ['admin', 'contenttypes', 'sessions', 'kpi']
            choices = []
            
            for model in apps.get_models():
                app_label = model._meta.app_label
                if app_label not in excluded_apps:
                    model_name = f"{app_label}.{model._meta.model_name}"
                    verbose_name = model._meta.verbose_name.title()
                    choices.append((model_name, f"{verbose_name}"))
            
            return sorted(choices)
        except (AppRegistryNotReady, OperationalError):
            return []
    
    def apply_condition(self, queryset, condition, target_field, target_value):
        if condition == 'EXACT':
            return queryset.filter(**{f"{target_field}": target_value})
        elif condition == 'CONTAINS':
            return queryset.filter(**{f"{target_field}__icontains": target_value})
        elif condition == 'NOT_EXACT':
            return queryset.exclude(**{f"{target_field}": target_value})
        elif condition == 'NOT_CONTAINS':
            return queryset.exclude(**{f"{target_field}__icontains": target_value})
        elif condition in ['GT', 'LT', 'GTE', 'LTE', 'EQUAL']:
            try:
                target_value = datetime.strptime(target_value, '%Y-%m-%d').date()
            except ValueError:
                pass
            if condition == 'EQUAL':
                return queryset.filter(**{f"{target_field}": target_value}) 
            else:
                lookup = {
                    'GT': 'gt',
                    'LT': 'lt',
                    'GTE': 'gte',
                    'LTE': 'lte',
                }[condition]
                return queryset.filter(**{f"{target_field}__{lookup}": target_value})
        elif condition == 'BETWEEN':
            from_value, to_value = target_value.split(' to ')
            try:
                from_value = datetime.strptime(from_value.strip(), '%Y-%m-%d').date()
                to_value = datetime.strptime(to_value.strip(), '%Y-%m-%d').date()
            except (ValueError, IndexError):
                pass
            return queryset.filter(**{f"{target_field}__range": (from_value, to_value)})

        return queryset


# Generated by Django 4.1.2 on 2025-01-12 16:54

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="KPI",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=100)),
                (
                    "description",
                    models.TextField(
                        blank=True, help_text="Optional description of the KPI"
                    ),
                ),
                ("model_field", models.CharField(max_length=100, verbose_name="Model")),
            ],
            options={
                "verbose_name": "KPI",
                "verbose_name_plural": "KPIs",
            },
        ),
        migrations.CreateModel(
            name="Card",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(help_text="Name of the card", max_length=100),
                ),
                (
                    "description",
                    models.TextField(
                        blank=True, help_text="Optional description of the card"
                    ),
                ),
                (
                    "icon",
                    models.CharField(help_text="Icon class or name", max_length=50),
                ),
                (
                    "value_suffix",
                    models.CharField(
                        blank=True,
                        help_text="Suffix to append to the value (e.g., %, $)",
                        max_length=50,
                    ),
                ),
                (
                    "operation",
                    models.CharField(
                        choices=[
                            ("count", "Count"),
                            ("count_distinct", "Count Distinct"),
                            ("sum", "Sum"),
                            ("avg", "Average"),
                            ("min", "Minimum"),
                            ("max", "Maximum"),
                        ],
                        default="count",
                        max_length=16,
                    ),
                ),
                (
                    "target_type",
                    models.CharField(
                        default="NUMBER",
                        help_text="Type of the target value",
                        max_length=20,
                    ),
                ),
                (
                    "target_field",
                    models.CharField(
                        blank=True,
                        help_text="Field to compare against target value",
                        max_length=100,
                    ),
                ),
                (
                    "condition",
                    models.CharField(
                        choices=[
                            (
                                "TEXT",
                                [
                                    ("EXACT", "Exactly matches"),
                                    ("CONTAINS", "Contains"),
                                    ("NOT_EXACT", "Does not exactly match"),
                                    ("NOT_CONTAINS", "Does not contain"),
                                ],
                            ),
                            (
                                "NUM",
                                [
                                    ("GT", "Greater Than"),
                                    ("LT", "Less Than"),
                                    ("EQ", "Equal"),
                                    ("GTE", "Greater Than or Equal"),
                                    ("LTE", "Less Than or Equal"),
                                    ("BETWEEN", "Between"),
                                ],
                            ),
                            ("NONE", "None"),
                        ],
                        default="EXACT",
                        max_length=20,
                    ),
                ),
                (
                    "target_value",
                    models.CharField(
                        blank=True,
                        help_text="Target value to achieve",
                        max_length=255,
                        null=True,
                    ),
                ),
                (
                    "kpi",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="card",
                        to="kpi.kpi",
                    ),
                ),
            ],
        ),
    ]



# views.py
from django.http import JsonResponse
from django.contrib.admin.views.decorators import staff_member_required
from django.views.decorators.http import require_GET
from django.apps import apps
from .models import KpiCard
@require_GET
@staff_member_required
def get_model_fields(request):
    if not request.user.is_authenticated or not request.user.is_staff:
        return JsonResponse({'error': 'Unauthorized'}, status=403)
        
    model_name = request.GET.get('model')
    try:
        app_label, model_name = model_name.split('.')
        model = apps.get_model(app_label, model_name)
        
        fields = []
        for field in model._meta.fields:
            # TODO: Handle relationship fields
            if field.is_relation:
                continue
            fields.append({
                'name': field.name,
                'verbose_name': str(field.verbose_name).title(),
                'type': field.get_internal_type()
            })
        
        return JsonResponse({'fields': fields})
    except Exception as e:
        return JsonResponse({'error': str(e)}, status=400)



@require_GET
@staff_member_required
def get_field_values(request):
    if not request.user.is_authenticated or not request.user.is_staff:
        return JsonResponse({'error': 'Unauthorized'}, status=403)
        
    model_name = request.GET.get('model')
    field_name = request.GET.get('field')
    
    try:
        app_label, model_name = model_name.split('.')
        model = apps.get_model(app_label, model_name)
                
        # Get distinct values
        values = list(model.objects.values_list(field_name, flat=True).distinct())
        
        # Convert all values to strings for consistent display
        values = [str(v) for v in values if v is not None]
        
        return JsonResponse({
            'values': values
        })
    except Exception as e:
        return JsonResponse({'error': str(e)}, status=400)


def dashboard_callback(request, context):
    cards = KpiCard.objects.all()
    context.update({"cards": cards})
    return context